%{
/* This is the lexical analyzer for the C* compiler.  
See the file README.TXT.
It takes the C* input and returns the next token for the parser.
The code is in lex, but the language used is C++ rather than C,
so this must be compiled with a C++ compiler after going through lex. */

	#include "parser.h"
	#include <cstring>
	#include <sstream>
	#include <climits>

	unsigned long commentLevel = 1;   /* Nest level of the current comment.
		Another integer that should go up to infinity, but doesn't */

	using namespace std;

	/* Increment the line #, which is used when an error
	 is reported so the user knows where errors occurred. */
	void inc_lineno() {
		if (lineno < ULONG_MAX)
			lineno++;
		else
			 compileError << "Program too big.\n";
	}

%}

digit [0-9]
letter [A-Za-z]
exponent [Ee][+-]?{digit}+
whitespace [ \t\v\f\r\n]

%x COMMENT
%x INCLUDE
%%

	/* Comments */
[/][*]	BEGIN COMMENT;

	/* Allow nested comments */
<COMMENT>[/][*]	{
		if (commentLevel < ULONG_MAX)
			commentLevel++;
		else
			compileError << linenum() << "Comment nested too deep.\n";
	}

<COMMENT>[*][/]	{
		if (commentLevel == 1)   //if we're in the top level comment, get out of the comment
			BEGIN INITIAL;
		else
			commentLevel--;
	}

	/* Newlines must be separate so the error checker will know the current line # */
<COMMENT,INITIAL,INCLUDE>[\n] inc_lineno();

<COMMENT>.

[/][/].*\n	inc_lineno();

	/* Other whitespace (spaces, tabs, etc.).
	Has to be ignored normally and when in an include statement.
	Without INCLUDE here, it would output the spaces beween the keyword
	"include" and the included file to stdout.*/
<INCLUDE,INITIAL>{whitespace}

	/* The range operator.
	E.g. the ".." in "int a[] = b[2..5];" 
	Remember to put it before the . operator so 2 of those don't get matched. */
[.][.]	return OP_RANGE;

	/* Floating point constants.  Slightly different than in C++; if a decimal
	point is present, at least one digit must follow. */
({digit}+{exponent})|(({digit}+)?[.]{digit}+{exponent}?)	{
		yylval.strVal = strdup(yytext);
		return FLOATCONST;
	}

	/* Integer constants.
	 * E.g.  the "1" in "int x = 1;" */
{digit}+ {
		yylval.intVal = atoi(yytext);
		return INTCONST;
	}

    /*Words that the users may not use, because the compiler
      needs them.*/
_clone  {
		compileError << linenum() << '`' << *yytext << "' is a reserved word.  Choose another identifier.\n";
}

	/*The keywords (reserved words).*/
for	return KEY_FOR;
select	return KEY_SELECT;
step	return KEY_STEP;
case	return KEY_CASE;
else	return KEY_ELSE;
function	return KEY_FUNCTION;
operator	return KEY_OPERATOR;
[=][>] return KEY_RETURNS;
return 	return KEY_RETURN;
break	return KEY_BREAK;
continue return KEY_CONTINUE;
goto	return KEY_GOTO;
if		return KEY_IF;
while	return KEY_WHILE;
do		return KEY_DO;
and     return KEY_AND;
[&][&]		return KEY_AND;
[|][|]		return KEY_OR;
or      return KEY_OR;
not     return KEY_NOT;
xor		return KEY_XOR;
enum	return KEY_ENUM;
cast    return KEY_CAST;
const	return KEY_CONST;
ref		return KEY_REF;
static  return KEY_STATIC;
null	return KEY_NULL;
true	return KEY_TRUE;
false	return KEY_FALSE;
new		return KEY_NEW;
delete	return KEY_DELETE;
destruct return KEY_DESTRUCT;
virtual return KEY_VIRTUAL;
try     return KEY_TRY;
catch   return KEY_CATCH;
throw   return KEY_THROW;
[.][.][.] return ELLIPSES;
friend  return KEY_FRIEND;

	/* C++ compatability */
cpp     return KEY_CPP;
include {
	BEGIN INCLUDE;
	return KEY_INCLUDE;
}
using   return KEY_USING;
namespace return KEY_NAMESPACE;
[:][:]	return COLON_COLON;

	/* The type specifiers.
	E.g. the "int" in "int x = 1;" 
	In a full language with classes, it would be better to load these
	into the symbol table as types, to simplify the parser. */
int	    return KEY_INT;
uint    return KEY_UINT;
double	return KEY_DOUBLE;
float   return KEY_FLOAT;
ldouble return KEY_LDOUBLE;
bool    return KEY_BOOL;
char    return KEY_CHAR;
thing	return KEY_THING;
type	return KEY_TYPE;

    /* Classes */
class   return KEY_CLASS;
public  return KEY_PUBLIC;
protected   return KEY_PROTECTED;
private return KEY_PRIVATE;
extends return KEY_EXTENDS;
final   return KEY_FINAL;
[<][:]  return SUBTYPE;
this    return KEY_THIS;

	/* Multiple character operators */
[:][=]	return '=';
[<][<]	return OUTPUT;
[>][>]	return INPUT;
[=][=]  return EQ;
[<][=]  return LE;
[>][=]  return GE;
[!][=]  return NE;
[~][=][=] return APPROX_EQUAL;
[+][+]  return INC;
[-][-]  return DEC;

	/* Shortcut assignment operators */
[+][=]  return PLUS_EQ;
[-][=]  return MINUS_EQ;
[*][=]  return TIMES_EQ;
[/][=]  return DIV_EQ;
[\\][=] return IDIV_EQ;
[%][=]  return MOD_EQ;
[&][=]	return BAND_EQ;
[|][=]	return BOR_EQ;
[>][>][=] return RSH_EQ;
[<][<][=] return LSH_EQ;

	/* C++ versions of boolean operators */
[!] return KEY_NOT;
[&&] return KEY_AND;
[||] return KEY_OR;

[*][*] return POWER;

	/* Single character operators */
[+*/\\%<>|&~?]  return *yytext;
[-]	return *yytext;

	/* Various special characters whose token value is the same as its character value */
[][.(){};,:@$]	return *yytext;

	/* Identifiers.
	 * E.g. the "x" in "int x := 1;" */
{letter}({letter}|{digit}|[_])* {
		yylval.strVal = strdup(yytext);
		return IDENT;
	}

	/* String literals */
<INCLUDE,INITIAL>["](([\\]["])|[^"])*["]	{
		yylval.strVal = strdup(yytext);
		BEGIN INITIAL;
		return STRING;
	}

<INCLUDE>[<][^<>]*[>] {
	yylval.strVal = strdup(yytext);
	BEGIN INITIAL;
	return STRING;
}

	/* Character literals */
['][\\]?.['] {
	yylval.strVal = strdup(yytext);
	return CHARCONST;
}
.	{
		compileError << linenum() << '`' << *yytext << "' is an invalid character.\n";
	}


%%
int yywrap() {
	return 1;
}

